/* CYGNUS LOCAL -- amylaar/calmrisc16 */
/* Definitions for CalmRISC16
   Copyright (C) 1997, 1998, 1999, 2000 Free Software Foundation, Inc.
   Contributed by Red Hat

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

#include <stdio.h>
#include "config.h"
#include "system.h"
#include "coretypes.h"
#include "tm.h"
#include "insn-config.h"
#include "rtl.h"
#include "tree.h"
#include "flags.h"
#include "expr.h"
#include "optabs.h"
#include "function.h"
#include "regs.h"
#include "hard-reg-set.h"
#include "output.h"
#include "insn-attr.h"
#include "toplev.h"
#include "recog.h"
#include "c-pragma.h"
#include "integrate.h"
#include "tm_p.h"
#include "target.h"
#include "target-def.h"
#include "real.h"
#include "langhooks.h"
#include "basic-block.h"
#include "ra.h"
#include "cfglayout.h"
#include "intl.h"
#include "ggc.h"
#include "conditions.h"
#include "insn-flags.h"
#include "output.h"
#include "obstack.h"
#include "except.h"


struct gcc_target targetm = TARGET_INITIALIZER;


static void output_stack_adjust (int, rtx, int, rtx (*)());
static int risc16_add_label_note (rtx *xp, void *data);

static rtx emit_frame_insn PARAMS ((rtx));
/* Define the information needed to generate branch and scc insns.  This is
   stored from the compare operation.  */
struct rtx_def *risc16_compare_op0;
struct rtx_def *risc16_compare_op1;

static void
output_stack_adjust (size, reg, temp, emit_fn)
     int size, temp;
     rtx reg;
     rtx (*emit_fn) ();
{
  if (size)
    {
      if (CONST_OK_FOR_M (size) || CONST_OK_FOR_N (size))
	(*emit_fn) (gen_addpsi_noclobber (reg, reg, GEN_INT (size)));
      else
	(*emit_fn) (gen_stack_adjust (reg, GEN_INT (size),
				      gen_rtx_REG (HImode, temp)));
    }
}

/* Returns the number of bytes offset between the frame pointer and the stack
   pointer for the current function.  SIZE is the number of bytes of space
   needed for local variables.  */
unsigned int
risc16_compute_frame_size (size)
     int size;
{
  int i;

  for (i = R0_REG; i <= LAST_ADDR_REG; i += GPR_GAP)
    if (SAVE_REG_P (i))
      size += i <= LAST_DATA_REG ? 2 : 4;
  return size + current_function_outgoing_args_size;
}

static rtx
emit_frame_insn (pattern)
     rtx pattern;
{
  rtx insn = emit_insn (pattern);

  RTX_FRAME_RELATED_P (insn) = 1;
  return insn;
}

/* Called after register allocation to add any instructions needed for the
   prologue.  Using a prologue insn is favored compared to putting all of the
   instructions in the FUNCTION_PROLOGUE macro, since it allows the scheduler
   to intermix instructions with the saves of the caller saved registers.  In
   some cases, it might be necessary to emit a barrier instruction as the last
   insn to prevent such scheduling.

   Also any insns generated here should have RTX_FRAME_RELATED_P(insn) = 1
   so that the debug info generation code can handle them properly.  */
void
risc16_expand_prologue ()
{
  rtx reg, last_reg;
  int pushed = 0;
  int i;
  int temp_reg = 4;
  HOST_WIDE_INT frame_size
    = get_frame_size () + current_function_outgoing_args_size;

  if (! CONST_OK_FOR_M (frame_size))
    regs_ever_live[temp_reg] = 1;
  for (last_reg = NULL_RTX, i = R0_REG; i <= LAST_DATA_REG; i += GPR_GAP)
    {
      if (regs_ever_live[i] && ! call_used_regs[i])
	{
	  reg = gen_rtx_REG (HImode, i);
	  pushed += 2;
	  if (last_reg)
	    {
	      emit_frame_insn (gen_push_rr (last_reg, reg));
	      last_reg = 0;
	    }
	  else
	    last_reg = reg;
	}
    }
  if (last_reg)
    emit_frame_insn (gen_push_r (last_reg));
  for (last_reg = NULL_RTX, i = FIRST_ADDR_REG; i <= LAST_ADDR_REG;
       i += GPR_GAP)
    {
      if (SAVE_REG_P (i))
	{
	  reg = gen_rtx_REG (PSImode, i);
	  pushed += 4;
	  if (last_reg)
	    {
	      emit_frame_insn (gen_push_aa (last_reg, reg));
	      last_reg = 0;
	    }
	  else
	    last_reg = reg;
	}
    }
  if (last_reg)
    emit_frame_insn (gen_push_a (last_reg));
  output_stack_adjust (-frame_size, stack_pointer_rtx, temp_reg,
		       &emit_frame_insn);
  if (frame_pointer_needed)
    {
      emit_insn (gen_movpsi (frame_pointer_rtx, stack_pointer_rtx));
      /* On the calmrisc16, push will write first to the
	 two bytes pointed to by sp.  So we have to add
	 two to current_function_outgoing_args_size.  */

      (emit_frame_insn
       (gen_addpsi3 (frame_pointer_rtx, frame_pointer_rtx,
		     GEN_INT (STACK_POINTER_OFFSET
			      + current_function_outgoing_args_size))));
    }
}

/* Called after register allocation to add any instructions needed for the
   epilogue.  Using a epilogue insn is favored compared to putting all of the
   instructions in the FUNCTION_PROLOGUE macro, since it allows the scheduler
   to intermix instructions with the saves of the caller saved registers.  In
   some cases, it might be necessary to emit a barrier instruction as the last
   insn to prevent such scheduling.  */

void
risc16_expand_epilogue ()
{
  rtx reg, last_reg;
  int pushed = 0;
  int i;
  int temp_reg = 4;
  HOST_WIDE_INT frame_size
    = get_frame_size () + current_function_outgoing_args_size;

  if (frame_pointer_needed)
    {
      /* Don't set sp from fp initially because it may corrupt stack
         during an interruption.  So don't use fp below in the
         function.  It has been changed. */
      frame_size -= STACK_POINTER_OFFSET + current_function_outgoing_args_size;
      output_stack_adjust (frame_size, frame_pointer_rtx, temp_reg,
			   &emit_insn);
      emit_insn (gen_movpsi (stack_pointer_rtx, frame_pointer_rtx));
    }
  else
    output_stack_adjust (frame_size, stack_pointer_rtx, temp_reg, &emit_insn);
  for (last_reg = NULL_RTX, i = LINK_REG; i >= FIRST_ADDR_REG; i -= GPR_GAP)
    {
      if (SAVE_REG_P (i))
	{
	  reg = gen_rtx_REG (PSImode, i);
	  pushed += 4;
	  if (last_reg)
	    {
	      emit_insn (gen_pop_aa (last_reg, reg));
	      last_reg = 0;
	    }
	  else
	    last_reg = reg;
	}
    }
  if (last_reg)
    emit_insn (gen_pop_a (last_reg));
  for (last_reg = NULL_RTX, i = R7_REG; i >= FIRST_DATA_REG; i -= GPR_GAP)
    {
      if (regs_ever_live[i] && ! call_used_regs[i])
	{
	  reg = gen_rtx_REG (HImode, i);
	  pushed += 2;
	  if (last_reg)
	    {
	      emit_insn (gen_pop_rr (last_reg, reg));
	      last_reg = 0;
	    }
	  else
	    last_reg = reg;
	}
    }
  if (last_reg)
    emit_insn (gen_pop_r (last_reg));
  emit_jump_insn (gen_return ());
}

void
risc16_emit_compare (code)
     enum rtx_code code;
{
  enum machine_mode mode = GET_MODE (risc16_compare_op0);

 
  if (mode == PSImode && code != EQ)
    {
      /* There are no PSImode inequality comparisons, so widen to SImode.  */
      rtx op0 = gen_reg_rtx (SImode);
      rtx op1 = gen_reg_rtx (SImode);

      emit_insn (gen_zero_extendpsisi2 (op0, risc16_compare_op0));
      risc16_compare_op0 = op0;
      if (GET_MODE (risc16_compare_op1) != VOIDmode)
	{
	  emit_insn (gen_zero_extendpsisi2 (op1, risc16_compare_op1));
	  risc16_compare_op1 = op1;
	}
      mode = SImode;
    }
  if (mode == SImode)
    emit_insn(gen_cmpsi_t_i(gen_rtx(code, PSImode),
			    risc16_compare_op0,
			    risc16_compare_op1
			    ));
  else
    emit_insn (gen_rtx_SET (VOIDmode, gen_rtx_REG (HImode, T_REG),
			    gen_rtx (code, HImode, risc16_compare_op0,
				     risc16_compare_op1)));

}

void
risc16_emit_scc (orig_code, reg_code, operands)
     enum rtx_code orig_code, reg_code;
     rtx *operands;
{
  if (arith_reg_operand (risc16_compare_op1, HImode))
    {
      rtx tmp = risc16_compare_op0;
      risc16_compare_op0 = risc16_compare_op1;
      risc16_compare_op1 = tmp;
      risc16_emit_compare (reg_code);
      emit_insn (gen_sbchic (operands[0], operands[0], operands[0]));
    }
  else
    {
      risc16_emit_compare (orig_code);
      emit_insn (gen_sbchic (operands[0], operands[0], operands[0]));
      emit_insn (gen_one_cmplhi2 (operands[0], operands[0]));
    }
}

/* Do any needed setup for a variadic function.  CUM has not been updated
   for the last named argument which has type TYPE and mode MODE.  */
void
generic_setup_incoming_varargs (cum, int_mode, type, pretend_size)
     CUMULATIVE_ARGS cum;
     int             int_mode;
     tree            type;
     int *           pretend_size;
{
}


int risc16_need_epilogue ()
{
  /* FIXME: revert to the old logic, using per-function information.
     see gccint, 13.4, Defining data structures for per-function information
 
     better yet: do as in the stormy16 backend. (or find even better ;)
  */

/*   int risc16_need_epilogue_known = 0; */
/*   // no longer static, hence the over-involved phrasing in the following -- Jyp */

/*   if (! risc16_need_epilogue_known) */
/*     { */
/*       rtx epilogue; */

/*       start_sequence (); */
/*       risc16_expand_epilogue (); */
/*       epilogue = get_insns (); */
/*       end_sequence (); */
/*       risc16_need_epilogue_known */
/* 	= (GET_CODE (epilogue) == SEQUENCE && XVECLEN (epilogue, 0) == 0 */
/* 	   ? -1 : 1); */
/*     } */
/*   return risc16_need_epilogue_known > 0; */

  int i;
  rtx reg, last_reg;

  if (frame_pointer_needed || 
      get_frame_size () + current_function_outgoing_args_size > 0)
    return 1;

  for (last_reg = NULL_RTX, i = LINK_REG; i >= FIRST_ADDR_REG; i -= GPR_GAP)
    if (SAVE_REG_P (i))
      return 1;

  for (last_reg = NULL_RTX, i = R7_REG; i >= FIRST_DATA_REG; i -= GPR_GAP)
    if (regs_ever_live[i] && ! call_used_regs[i])
      return 1;
  
  return 0;
}

#if 0
REMOVED: hook no longer exists
void
risc16_function_epilogue ()
{
  risc16_need_epilogue_known = 0;
}
#endif

/* Print a memory address as an operand to reference that memory location.  */
void
risc16_print_operand_address (stream, address)
     FILE *stream;
     rtx address;
{
  rtx offset = const0_rtx;

  if (GET_CODE (address) == PLUS)
    {
      offset = XEXP (address, 1);
      address = XEXP (address, 0);
    }
  if (GET_CODE (address) != REG || REGNO (address) < FIRST_ADDR_REG)
    abort ();
  if (GET_CODE (offset) == CONST_INT)
    fprintf (stream, "@[a%d+%ld]", REGNO (address) / GPR_GAP, INTVAL (offset));
  else if (GET_CODE (offset) == REG)
    fprintf (stream, "@[a%d+r%d]", REGNO (address) / GPR_GAP, REGNO (offset));
  else
    abort ();
}

/* Print an operand to a assembler instruction.

  'E'  print the extension part of a 22 bit address register.
  'A'  print least significant part of absolute operand (if const_int)
  'B'  print 2nd least significant part of absolute operand (if const_int)
  'O'  print a constant without the #
  'R'  print least significant part of the operand
  'S'  print 2nd least significant part of the operand
  'T'  print 3rd least significant part of the operand
  'U'  print 4st least significant part of the operand
  'N'  print raw register number
  'M'  print negated value
  'b'  print add / sub depending on sign of operand
  'd'  print decc / incc depending on sign of operand
  'i'  print incc / decc depending on sign of operand
  'L'  print ld / ldb / ldw depending on operand
  'o'  print an operator
  's'  print sub / add depending on sign of operand
  't'  print sbc / adc depending on sign of operand
  ' '  print \n\t
*/
void
risc16_print_operand (file, x, code)
     FILE * file;
     rtx    x;
     int    code;
{
  switch (code)
    {
    case 'E':
      fprintf (file, "e%d", REGNO (x) / GPR_GAP);
      return;
    case 'O':
      output_addr_const (file, x);
      return;
    case 'R': case 'S': case 'T': case 'U':
      {
	int word = code - 'R';

	if (GET_CODE (x) == REG)
	  {
	    int offset;

	    switch (GET_MODE (x))
	      {
	      case PSImode:
	      case HImode:
		offset = 0;
		break;
	      case SImode:
	      case SFmode:
		offset = 1;
		break;
	      case DImode:
	      case DFmode:
		offset = 3;
		break;
	      default:
		offset = INT_MAX;
	      }
	    fputs (reg_names[REGNO (x) + offset - word], file);
	  }
	else if (GET_CODE (x) == CONST_INT)
	  fprintf (file, "#%ld", (long) (INTVAL (x) >> word * 16) & 0xffff);
	else if (GET_CODE (x) == CONST_DOUBLE)
	  {
	    if (word < 2)
	      fprintf (file, "#%ld",
		       (long) (CONST_DOUBLE_LOW (x) >> word * 16) & 0xffff);
	    else
	      fprintf (file, "#%ld",
		       ((long) (CONST_DOUBLE_HIGH (x) >> (word - 2) * 16)
			& 0xffff));
	  }
	else if (CONSTANT_P (x) && word < 2)
	  {
	    if (word == 0
		&& ((GET_CODE (x) == CONST && GET_CODE (XEXP (x, 0)) == LO_SUM)
		    || GET_CODE (x) == HIGH))
	      break;
	    fputs (word ? "#%hi(" : "#%lo(", file);
	    output_addr_const (file, x);
	    fputs (")", file);
	  }
	else if (GET_CODE (x) == MEM)
	  {
	    x = adjust_address(x, HImode, (GET_MODE_SIZE (GET_MODE (x))
					   - 2 - word * 2));
	    risc16_print_operand_address (file, XEXP (x, 0));
	  }
	else
	  abort ();
	return;
      }
    case 'A': case 'B':
      {
	int word = code - 'A';

	if (GET_CODE (x) == CONST_INT)
	  {
	    HOST_WIDE_INT val = INTVAL (x);
	    if (val < 0)
	      val = -val;
	    fprintf (file, "#%ld", (long) (val >> word * 16) & 0xffff);
	  }
	else if (GET_CODE (x) == CONST_DOUBLE)
	  abort ();
	else
	  risc16_print_operand (file, x, 'R' + word);
	return;
      }
    case 'N':
      fprintf (file, "%d", REGNO (x));
      return;
    case 'M':
      if (GET_CODE (x) == CONST_INT)
	{
	  risc16_print_operand (file, GEN_INT (-INTVAL (x)), 'R');
	  return;
	}
      else if (CONSTANT_P (x) && GET_MODE (x) == HImode)
	{
	  if (GET_CODE (x) == CONST)
	    x = XEXP (x, 0);
	  x = gen_rtx_CONST (HImode, gen_rtx_NEG (HImode, x));
	  break;
	}
      else
	abort ();
    case 'b':
      fputs (GET_CODE (x) == CONST_INT && INTVAL (x) < 0 ? "sub" : "add", file);
      return;
    case 's':
      fputs (GET_CODE (x) == CONST_INT && INTVAL (x) < 0 ? "add" : "sub", file);
      return;
    case 't':
      fputs (GET_CODE (x) == CONST_INT && INTVAL (x) < 0 ? "adc" : "sbc", file);
      return;
    case 'i':
      fputs (GET_CODE (x) == CONST_INT && INTVAL (x) < 0 ? "decc" : "incc",
	     file);
      return;
    case 'L':
      if (GET_CODE (x) == MEM
	  || (GET_CODE (x) == SUBREG && GET_CODE (SUBREG_REG (x)) == MEM)
	  || (GET_CODE (x) == ZERO_EXTRACT && GET_CODE (SUBREG_REG (x)) == MEM))
	fputs (GET_MODE (x) == QImode ? "ldb" : "ldw", file);
      else
	fputs ("ld", file);
      return;
    case 'd':
      fputs (GET_CODE (x) == CONST_INT && INTVAL (x) < 0 ? "incc" : "decc",
	     file);
      return;
    case 'o':
      switch (GET_CODE (x))
	{
	case GE:  fputs ("ge",  file); break;
	case GT:  fputs ("gt",  file); break;
	case GEU: fputs ("uge", file); break;
	case GTU: fputs ("ugt", file); break;
	case EQ:  fputs ("eq",  file); break;
	default: abort ();
	}
      return;
    case ' ':
      fputs ("\n\t", file);
      return;
    case 0:
      /* Handled below.  */
      break;
      
    default:
      output_operand_lossage ("generic_print_operand: unknown code");
      return;
    }

  switch (GET_CODE (x))
    {
    case REG:
      if (GET_MODE (x) == PSImode
	  && REGNO (x) >= FIRST_ADDR_REG && REGNO (x) <= LAST_ADDR_REG)
	{
	  if (REGNO (x) == STACK_POINTER_REGNUM)
	    fputs ("sp", file);
	  else
	    fprintf (file, "a%d", REGNO (x) / GPR_GAP);
	}
      else
	fputs (reg_names [REGNO (x)], file);
      break;

    case ZERO_EXTRACT:
      if (! risc16_movhisrc_operand (x, VOIDmode))
	abort ();
      if (GET_CODE (XEXP (x, 0)) == REG)
	fprintf (file, "e%d", REGNO (XEXP (x, 0)));
      else if (GET_CODE (XEXP (x, 0)) == MEM)
	risc16_print_operand (file, XEXP (x, 0), code);
      else
	abort ();
      break;

    case MEM:
      output_address (XEXP (x, 0));
      break;

    case HIGH:
      fputs ("#%hi(", file);
      output_addr_const (file, XEXP (x, 0));
      fputs (")", file);
      break;
    case CONST:
      if (GET_CODE (XEXP (x, 0)) == LO_SUM)
	{
	  fputs ("#%lo(", file);
	  output_addr_const (file, XEXP (XEXP (x, 0), 0));
	  fputs (")", file);
	  break;
	}
      /* Fall through.  */
      
    default:
      fputs ("#", file);
      output_addr_const (file, x);
      break;
    }

  return;
}

const char *
output_branch (names, insn, operands, revcond_name)
     const char *names, *revcond_name;
     rtx insn;
     rtx *operands;
{
  const char *trailer = "\n";
  /* The call to final_scan_insn will clobber operands.  */
  rtx target = operands[0];
  int no_delay_insn
    = (! final_sequence
       || INSN_ANNULLED_BRANCH_P (XVECEXP (final_sequence, 0, 0)));

  /* Note that if this is a jump with the delay slot filled, 'insn' will
     actually be the inner jump_insn, thus the length of the delay slot
     insn is not included.  Thus, the delay slot length is only counted
     in if it remains unfilled, which can only happen for a bsrd.
     (output_branch is not used for the bnzd machine insn.)
     When we are passed a bsrd insn that hasn't got its delay slot filled,
     we just output it as a jsr.  This takes the same number of bytes and
     cycles as a bsrd;nop. */
  if (get_attr_length (insn) >= 4)
    {
      /* This is the long version of the branch, which has no delay slot.  */
      if (! no_delay_insn)
	{
	  /* If this is a condbranch and the delay slot insn sets the T
	     register, we must reverse the branch and branch around a jump.  */
	  if (revcond_name
	      && get_attr_type (XVECEXP (final_sequence, 0, 1)) == TYPE_T_SET)
	    {
	      fprintf (asm_out_file, "\t%s\t0f\n", revcond_name);
	      names = "bra\0jmp";
	      trailer = "\n0:\n";
	    }
	  /* Output the delay slot insn now.  */
	  final_scan_insn (XVECEXP (final_sequence, 0, 1), asm_out_file,
			   optimize, 0, 1, NULL);
	  INSN_DELETED_P (XVECEXP (final_sequence, 0, 1)) = 1;
	}
      fprintf (asm_out_file, "\t%s\t", names+4);
    }
  else
    fprintf (asm_out_file, "\t%s%s\t", names, "d" + no_delay_insn);
  PRINT_OPERAND (asm_out_file, target, 'O');
  fputs (trailer, asm_out_file);
  return "";
}

int
jump_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  if (GET_MODE (op) != mode && mode != VOIDmode)
    return 0;
  if (CONSTANT_P (op))
    return 1;
  return register_operand (op, mode);
}

int
jump_mem_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  if (GET_MODE (op) != mode && mode != VOIDmode)
    return 0;
  if (GET_CODE (op) != MEM)
    return 0;
  return jump_operand (XEXP (op, 0), PSImode);
}

int
const_1_operand (op, mode)
     rtx op;
     enum machine_mode mode ATTRIBUTE_UNUSED;
{
  return GET_CODE (op) == CONST_INT && INTVAL (op) == 1;
}

/* Returns 1 if OP is a normal arithmetic register.  */

int
arith_reg_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  int regno;

  if (! register_operand (op, mode))
    return 0;
  if (GET_CODE (op) == REG)
    regno = REGNO (op);
  else if (GET_CODE (op) == SUBREG && GET_CODE (SUBREG_REG (op)) == REG)
    regno = REGNO (SUBREG_REG (op));
  else
    /* We allow (SUBREG (MEM ...)) before reload.  */
    return 1;

  return regno != T_REG;
}

/* Returns 1 if OP is a data register.  */

int
data_reg_operand (op, mode)
     rtx op;
     enum machine_mode mode;
{
  int regno;

  if (! register_operand (op, mode))
    return 0;
  if (GET_CODE (op) == REG)
    regno = REGNO (op);
  else if (GET_CODE (op) == SUBREG && GET_CODE (SUBREG_REG (op)) == REG)
    regno = subreg_regno (op);
    /* Used to be    regno = REGNO (SUBREG_REG (op)) + SUBREG_WORD (op); -- Jyp */
  else
    /* We allow (SUBREG (MEM ...)) before reload.  */
    return 1;

  return regno <= LAST_DATA_REG;
}

int
risc16_movsrc_operand (op, mode)
    rtx op;
    enum machine_mode mode;
{
  if (general_operand (op, mode))
    {
      if (GET_CODE (op) == SUBREG)
	{
	  op = SUBREG_REG (op);
	  if (mode == PSImode && GET_MODE (op) != PSImode)
	    return 0;
	}
      else if (reload_in_progress || reload_completed)
	return 1;
      if ((GET_CODE (op) == REG
	   && REGNO (op) >= FIRST_VIRTUAL_REGISTER
	   && REGNO (op) <= LAST_VIRTUAL_REGISTER)
	  || (op == frame_pointer_rtx && ! reload_completed)
	  || op == arg_pointer_rtx)
	return 0;
      return 1;
    }
  return 0;
}

int
risc16_movhisrc_operand (op, mode)
    rtx op;
    enum machine_mode mode;
{
  rtx loc, size, pos;

  if (risc16_movsrc_operand (op, mode))
    return 1;
  if (GET_CODE (op) != ZERO_EXTRACT)
    return 0;
  if (GET_MODE (op) != mode && mode != VOIDmode)
    return 0;
  loc = XEXP (op, 0);
  if (GET_CODE (loc) == REG)
    {
      if (GET_MODE (loc) != PSImode
	  || ((REGNO (loc) < FIRST_ADDR_REG || REGNO (loc) > LAST_ADDR_REG)
	      && REGNO (loc) < FIRST_PSEUDO_REGISTER))
	return 0;
    }
  else if (GET_CODE (loc) == MEM)
    {
      if (! reload_completed || ! memory_operand (loc, PSImode))
	return 0;
    }
  else
    return 0;
  size = XEXP (op, 1);
  if (GET_CODE (size) != CONST_INT || INTVAL (size) != 6)
    return 0;
  pos = XEXP (op, 2);
  if (GET_CODE (pos) != CONST_INT || INTVAL (pos) != 16)
    return 0;
  return 1;
}

int
risc16_movdst_operand (op, mode)
    rtx op;
    enum machine_mode mode;
{
  if (nonimmediate_operand (op, mode))
    {
      if (GET_CODE (op) == SUBREG)
	{
	  op = SUBREG_REG (op);
	  if (mode == PSImode && GET_MODE (op) != PSImode)
	    return 0;
	}
      return 1;
    }
  return 0;
}

int
risc16_movhidst_operand (op, mode)
    rtx op;
    enum machine_mode mode;
{
  return risc16_movhisrc_operand (op, mode) && ! CONSTANT_P (op);
}

int
risc16_greater_operator (op, mode)
    rtx op;
    enum machine_mode mode;
{
  enum rtx_code code = GET_CODE (op);

  return ((mode == VOIDmode || GET_MODE (op) == mode)
	  && (code == GE || code == GT || code == GEU || code == GTU));
}

int
risc16_move_check (operands)
     rtx *operands;
{
  rtx op0 = operands[0];
  rtx op1 = operands[1];

  if (GET_CODE (op0) == SUBREG)
    op0 = SUBREG_REG (op0);
  if (GET_CODE (op1) == SUBREG)
    op1 = SUBREG_REG (op1);
  if (GET_CODE (op0) == MEM && GET_CODE (op1) == MEM)
    return 0;
  /* Check if machine_dependent_reorg already fixed up Z0 / Z1 clobbers.  */
  if (0)
    {
      if (GET_CODE (op0) == REG && LOOP_COUNT_REGNUM_P (REGNO (op0))
	  && (GET_CODE (operands[2]) != REG
	      || REGNO (operands[2]) != LOOP_COUNT_COND (REGNO (op0))))
	return 0;
    }
  return 1;
}
#if 0
int
risc16_arith_check (operands)
     rtx *operands;
{
  rtx op0 = operands[0];
  rtx op1 = operands[1];

  /* Check if machine_dependent_reorg already fixed up Z0 / Z1 clobbers.  */
  if (0)
    {
      if (GET_CODE (op0) == REG && LOOP_COUNT_REGNUM_P (REGNO (op0))
	  && (GET_CODE (operands[2]) != REG
	      || REGNO (operands[2]) != LOOP_COUNT_COND (REGNO (op0))))
	return 0;
    }
  return 1;
}
#endif
int
risc16_address_cost (addr, mode)
     rtx  addr;
     enum machine_mode mode;
{
  rtx base, index;
  unsigned HOST_WIDE_INT offset;

  if (GET_CODE (addr) != PLUS)
    return 1;
  index = XEXP (addr, 1);
  if (GET_CODE (index) == REG)
    return 1;
  if (GET_CODE (index) != CONST_INT)
    return 2;
  offset = INTVAL (index);
  base = XEXP (addr, 0);
  if (GET_CODE (base) != REG)
    return 2;
  if (mode == PSImode)
    return (offset & 1) == 0 && offset < 32 ? 1 : 2;
  else if (mode == QImode)
    return offset < 16 ? 1 : 2;
  else
    {
      offset += GET_MODE_SIZE (mode);
      if ((offset & 1) == 0 && offset <= 32)
	return 1;
      if (REGNO (base) == STACK_POINTER_REGNUM
	  && (offset & 1) == 0 && offset <= 512)
	return 1;
    }
  return 2;
}

#ifndef IS_ASM_LOGICAL_LINE_SEPARATOR
#define IS_ASM_LOGICAL_LINE_SEPARATOR(C) ((C) == ';')
#endif

int
risc16_insn_length_adjustment (insn, orig_len)
     rtx insn;
     int orig_len;
{
  /* Instructions with unfilled delay slots take up an extra two bytes for
     the nop in the delay slot.  */
  if (((GET_CODE (insn) == INSN && GET_CODE (PATTERN (insn)) == PARALLEL)
       || GET_CODE (insn) == CALL_INSN)
      && get_attr_needs_delay_slot (insn) == NEEDS_DELAY_SLOT_YES
      /* The delay slot is only mandatory for the shortest form.  */
      && orig_len == 2)
    return 2;

  /* Check if this is an asm insn.  */
  if (GET_CODE (insn) == INSN)
    {
      int sum = 0;
      rtx body = PATTERN (insn);
      const char *template;
      char c;
      int maybe_label = 1;

      if (GET_CODE (body) == ASM_INPUT)
	template = XSTR (body, 0);
      else if (asm_noperands (body) >= 0)
	template
	  = decode_asm_operands (body, NULL, NULL, NULL, NULL);
      else
	return 0;
      do
	{
	  int asm_adjust = 0;

	  do
	    c = *template++;
	  while (c == ' ' || c == '\t');
	  /* A jpf / jpt / jmp / jsr needs an extra 2 bytes, unless
	     the argument is an address register.  */
	  if ((c == 'j' || c == 'J')
	      && (template[4] != 'a' && template[4] != 'A'))
	    asm_adjust = 2;
	  while (c && c != '\n' && ! IS_ASM_LOGICAL_LINE_SEPARATOR (c))
	    {
	      /* If this is a label, it is obviously not a long insn.  */
	      if (c == ':' && maybe_label)
		{
		  asm_adjust = 0;
		  break;
		}
	      else if (c == '\'' || c == '"')
		maybe_label = 0;
	      /* We pretend here that every immediate and memory access needs
		 an extra two bytes.  This could be refined if it appears
		 important enough.  */
	      if (c == '#' || c == '@')
		asm_adjust = 2;
	      c = *template++;
	    }
	  sum += asm_adjust;
	  maybe_label = c != ':';
	}
      while (c);
      return sum;
    }
  return 0;
}

#define GEN_INT_HIMODE(x) (gen_int_mode(x, HImode))

rtx
risc16_split_operand (op, size, word)
     rtx op;
     int size, word;
{
  if (size == 1 && word == 0)
    return op;
  if (GET_CODE (op) == REG)
    {
      if (GET_MODE (op) != PSImode || REGNO (op) <= LAST_DATA_REG)
	return gen_rtx_REG (HImode, REGNO (op) + word);
      if (word)
        return gen_rtx_REG (HImode, REGNO (op));
      return gen_rtx_ZERO_EXTRACT (HImode, op, GEN_INT (6), GEN_INT (16));
    }
  if (GET_CODE (op) == CONST_INT)
    {
      int shift = (size - 1 - word) * 16;

      if (shift >= HOST_BITS_PER_WIDE_INT)
	return INTVAL (op) < 0 ? GEN_INT_HIMODE (0xffff) : const0_rtx;
      return GEN_INT_HIMODE ((INTVAL (op) >> shift) & 0xffff);
    }
  if (GET_CODE (op) == CONST_DOUBLE)
    {
      op = operand_subword (op, word, 0,
			    (GET_MODE (op) == VOIDmode
			     ? (size == 2 ? SImode : DImode) : GET_MODE (op)));
      if (INTVAL (op) < 0)
	op = GEN_INT_HIMODE (INTVAL (op) & 0xffff);
      return op;
    }
  if (CONSTANT_P (op) && size <= 2)
    return (word
	    ? gen_rtx_CONST (HImode, gen_rtx_LO_SUM (HImode, op, pc_rtx))
	    : gen_rtx_HIGH (HImode, op));
  if (GET_CODE (op) == MEM)
    return change_address (op, HImode, plus_constant (XEXP (op, 0), 2 * word));
  abort ();
}

void
risc16_split_move (dst, src)
     rtx dst, src;
{
  int size = GET_MODE_SIZE (GET_MODE (dst)) / 2;
  int i, step, end;
  rtx last_insn = NULL_RTX;

  if (GET_CODE (src) == REG && GET_CODE (dst) == REG
      && REGNO (dst) < REGNO (src))
    {
      i = 0;
      step = 1;
      end = size;
    }
  else
    {
      i = size - 1;
      step = -1;
      end = -1;
    }
  for (; i != end; i += step)
    {
      rtx newdst, newsrc, insn;

      newdst = risc16_split_operand (dst, size, i);
      newsrc = risc16_split_operand (src, size, i);
      insn = gen_movhi_i (newdst, newsrc);
      /* If we are setting a loop counter register,
	 insert the associated flag register in the clobber.  */
      if (GET_CODE (newdst) == REG && LOOP_COUNT_REGNUM_P (REGNO (newdst)))
	XEXP (XVECEXP (insn, 0, 1), 0)
	  = gen_rtx_REG (PQImode, LOOP_COUNT_COND (REGNO (newdst)));

      if (GET_CODE (src) != REG && reg_overlap_mentioned_p (newdst, src))
	{
	  if (last_insn)
	    abort ();
	  last_insn = insn;
	}
      else
	emit_insn (insn);
    }
  if (last_insn)
    emit_insn (last_insn);
}

void
risc16_split_logical_operation (code, dst, src)
     enum rtx_code code;
     rtx dst, src;
{
  int size = GET_MODE_SIZE (GET_MODE (dst)) / 2;
  int i;
  rtx insn = NULL_RTX;

  for (i = size - 1; i >= 0; i--)
    {
      rtx newsrc = risc16_split_operand (src, size, i);
      rtx newdst, clobber, t_clobber;

      /* If optimizing, and we'd emit a no-op insn, continue.  */
      if (optimize && GET_CODE (newsrc) == CONST_INT
	  && ((code == XOR || code == IOR) && INTVAL (newsrc) == 0
	      || code == AND && INTVAL (newsrc) == 65535))
	continue;

      newdst = risc16_split_operand (dst, size, i);
      if (GET_CODE (newdst) == REG && LOOP_COUNT_REGNUM_P (REGNO ((newdst))))
	clobber = gen_rtx_REG (PQImode, LOOP_COUNT_COND (REGNO (newdst)));
      else
	clobber = gen_rtx_SCRATCH (PQImode);
      clobber = gen_rtx_CLOBBER (VOIDmode, clobber);
      
      /* If optimizing, replace logical operation with SET where appropriate.
	 This allows later passes to delete dead feeding insns.  */
      if (optimize && GET_CODE (newsrc) == CONST_INT
	  && ((code == IOR && INTVAL (newsrc) == 65535)
	      || (code == AND && INTVAL (newsrc) == 0)))
	{
	  insn = gen_rtx_SET (VOIDmode, newdst, newsrc);
	  insn = gen_rtx_PARALLEL (VOIDmode, gen_rtvec (2, insn, clobber));
	}
      else
	{
	  newsrc = gen_rtx_fmt_ee (code, HImode, newdst, newsrc);
	  insn = gen_rtx_SET (VOIDmode, newdst, newsrc);
	  t_clobber = gen_rtx_CLOBBER (VOIDmode, gen_rtx_REG (HImode, T_REG));
	  insn = gen_rtx_PARALLEL (VOIDmode,
				   gen_rtvec (3, insn, clobber, t_clobber));
	}
      emit_insn (insn);
    }

  /* If we find that we don't actually need any insn, emit a no-op move.  */
  if (! insn)
    risc16_split_move (dst, dst);
}

void
emit_hi_insn (pat)
     rtx pat;
{
  rtx dst = SET_DEST (XVECEXP (pat, 0, 0));

  if (GET_CODE (dst) == REG && LOOP_COUNT_REGNUM_P (REGNO (dst)))
    XEXP (XVECEXP (pat, 0, XVECLEN (pat, 0) - 1), 0)
      = gen_rtx_REG (PQImode, LOOP_COUNT_COND (REGNO (dst)));
  emit_insn (pat);
}

static int
risc16_add_label_note (xp, data)
     rtx *xp;
     void *data;
{
  rtx x = *xp;
  rtx insn = data;

  if (GET_CODE (x) == LABEL_REF)
    {
      REG_NOTES (insn)
	= gen_rtx_EXPR_LIST (REG_LABEL, XEXP (x, 0), REG_NOTES (insn));
      LABEL_NUSES (XEXP (x, 0))++;
    }
  return 0;
}

rtx
emit_label_insn (insn, maybe_labels)
     rtx insn, maybe_labels;
{
  insn = emit_insn (insn);
  for_each_rtx (&maybe_labels, risc16_add_label_note, insn);
  return insn;
}

#if 0
Hook no longer exists. Plus, this is total crap. -- Jyp


int split_dbra_p;

/* DBRA_INSN is an decrement_and_branch_until_zero insn.
   SETTER is the preceding movhi_i / addhi3 / subhi3 to set
   the register that is being decremented,
   We expect most opportunities to use bnzd to be of the form
     ld rn,X
   loop_start:
     loop body
     bnzd rn,loop_start
  where rn is the register with the number REGNO.
  So we search backwards from the dbra to the previous insn to set rn.

  We must make sure that there is no label inside the loop that is jumped
  to from outside the loop.

  Are you out of your mind, ffs?! -- Jyp

  LIFE_NUM is the ordinal number of the lifetime
  of rn between SETTER and DBRA_INSN among the lifetimes of rn in the
  current function.  We have prepared LABEL_LIFE to map each
  CODE_LABEL_NUMBER to the ordinal number of the lifetime of rn it falls into.
  So in order to check that there is no jump into the loop from outside,
  sum up all the LABEL_NUSES of labels between SETTER and DBRA_INSN,
  and check that we find a matching number of LABEL_REFS to labels in this
  lifetime of rn.

  We may have lumped several lifetimes into one, so we still have to check
  that rn isn't clobbered between SETTER and DBRA_INSN.  */

static void
try_fixup_dbra (setter, dbra_insn, life_num, label_life)
     rtx setter, dbra_insn;
     int life_num, *label_life;
{
  rtx set_src, reg, flag_reg, insn;
  int unknown_references;
  int need_flag_set = 0;
  int insn_count;
  rtx (*gfun) PARAMS ((rtx, rtx, rtx));

  if (! setter)
    return;
  switch (recog_memoized (setter))
    {
    case CODE_FOR_movhi_i:
      gfun = gen_set_loopcount;
      if (memory_operand (set_src, VOIDmode))
	{
	  if (optimize_size)
	    return;
	  need_flag_set = 1;
	}
      break;
    case CODE_FOR_addhi3:
    case CODE_FOR_subhi3:
      gfun = gen_calc_loopcount;
      break;
    default:
      return;
    }
  set_src = SET_SRC (XVECEXP (PATTERN (setter), 0, 0));
  reg = SET_DEST (XVECEXP (PATTERN (setter), 0, 0));
  flag_reg = XEXP (XVECEXP (PATTERN (dbra_insn), 0, 2), 0);

  /* If the loop start label is not between loop count setting insn and
     the dbra, the setting insn is likely inside the loop (e.g. created
     by reload).  If we have to insert another insn there, bnzd is not
     worth while.  */
  if (label_life[CODE_LABEL_NUMBER (JUMP_LABEL (dbra_insn))] != life_num)
    {
      if (need_flag_set)
	return;
      unknown_references = 0;
    }
  else
    unknown_references = -1;

  /* bnzd is only suitable for loops with up to 64 insns, so if we see that
     we have more, don't bother to scan any further.  This can not only speed
     up compilation, it will also allow delayed branch scheduling to split
     this dbra, possibly putting the sub in some delay slot.  */
  insn_count = 64;
  for (insn = PREV_INSN (dbra_insn); insn != setter; insn = PREV_INSN (insn))
    {
      if (GET_CODE (insn) == CODE_LABEL)
	{
	  unknown_references += LABEL_NUSES (insn);
	  /* If this is the label that the dbra references, any preceding
	     insn doesn't contribute to the branch offset, so disable the
	     early exit for offset overflow.  */
	  if (JUMP_LABEL (dbra_insn) == insn)
	    insn_count = 0;
	}
      else if (INSN_P (insn))
	{
	  if (! --insn_count)
	    return;
	  if (reg_set_p (reg, insn) || reg_set_p (flag_reg, insn))
	    return;
	  if (GET_CODE (insn) == JUMP_INSN
	      && JUMP_LABEL (insn)
	      && label_life [CODE_LABEL_NUMBER (JUMP_LABEL (insn))] == life_num)
	    unknown_references--;
	}
    }
  if (unknown_references)
    return;

  PATTERN (dbra_insn) = gen_bnzd (JUMP_LABEL (dbra_insn), reg, flag_reg);
  INSN_CODE (dbra_insn) = CODE_FOR_bnzd;
  if (need_flag_set)
    emit_insn_after ((*gfun) (reg, set_src, flag_reg), setter);
  else
    {
      PATTERN (setter) = (*gfun) (reg, set_src, flag_reg);
      INSN_CODE (setter) = -1;
    }
}

void
risc16_machine_dependent_reorg (first)
     rtx first;
{
  rtx insn;
  rtx loop_count_setter[GPR_GAP + 1];
  int count_life[GPR_GAP + 1];
  int *label_life[GPR_GAP + 1];
  int label_life_size;

  split_dbra_p = 1;

  /* ??? Some deleted EH handlers remain referenced.  */
  for (insn = first; insn; insn = NEXT_INSN (insn))
    if (GET_CODE (insn) == INSN
	&& GET_CODE (PATTERN (insn)) == SET
	&& GET_CODE (SET_SRC (PATTERN (insn))) == LABEL_REF
	&& (GET_CODE (XEXP (SET_SRC (PATTERN (insn)), 0)) != CODE_LABEL
	    || (! LABEL_PRESERVE_P (XEXP (SET_SRC (PATTERN (insn)), 0))
		&& NEXT_INSN (XEXP (SET_SRC (PATTERN (insn)), 0)) == NULL_RTX)))
    SET_SRC (PATTERN (insn)) = const0_rtx;

  if (! optimize)
    return;
  label_life_size = max_label_num () * sizeof *label_life[0];
  label_life[0] = (int *) alloca (label_life_size);
  label_life[GPR_GAP] = (int *) alloca (label_life_size);
  count_life[0] = 0;
  count_life[GPR_GAP] = 0;
  for (insn = first; insn; insn = NEXT_INSN (insn))
    {
      if (GET_CODE (insn) == CODE_LABEL)
	{
	  label_life[0][CODE_LABEL_NUMBER (insn)] = count_life[0];
	  label_life[GPR_GAP][CODE_LABEL_NUMBER (insn)] = count_life[GPR_GAP];
	}
      else if (GET_CODE (insn) == JUMP_INSN
	       && GET_CODE (PATTERN (insn)) == SET
	       && recog_memoized (insn) == CODE_FOR_dbra_i)
	{
	  int n = true_regnum (SET_DEST (XVECEXP (PATTERN (insn), 0, 1)));
	  count_life[n - LOOP_COUNT0]++;
	}
      else if (GET_CODE (insn) == INSN
	       && GET_CODE (PATTERN (insn)) == PARALLEL
	       && get_attr_loop_count_set (insn))
	{
	  int n = true_regnum (SET_DEST (XVECEXP (PATTERN (insn), 0, 0)));
	  if (LOOP_COUNT_REGNUM_P (n))
	    {
	      int i = 1;
	      rtx clobber;

	      count_life[n - LOOP_COUNT0]++;

	      /* If we don't have a proper clobber, add one.  */
	      do
		clobber = XVECEXP (PATTERN (insn), 0, i), i++;
	      while (GET_CODE (clobber) != CLOBBER);
	      if (GET_CODE (XEXP (clobber, 0)) == REG)
		{
		  /* Register allocation might leave spurious clobbers,
		     but not for fixed registers.  */
		  if (fixed_regs[REGNO (XEXP (clobber, 0))])
		    abort ();
		}
	      else if (GET_CODE (XEXP (clobber, 0)) != SCRATCH)
		abort ();

	      n = LOOP_COUNT_COND (n);
	      if (GET_CODE (XEXP (clobber, 0)) != REG
		  || REGNO (XEXP (clobber, 0)) != n)
		XEXP (clobber, 0) = gen_rtx_REG (PQImode, n);
	    }
	}
      /* If this is a leaf function, return_address_pointer_rtx couldn't be
	 eliminated.  Replace the memory reference with the link register.  */
      else if (GET_CODE (insn) == INSN
	       && GET_CODE (PATTERN (insn)) == SET)
	{
	  rtx pat = PATTERN (insn);
	  rtx src = SET_SRC (pat);

	  if (src == return_address_pointer_rtx)
	    abort ();
	  if (GET_CODE (src) == MEM
	      && XEXP (src, 0) == return_address_pointer_rtx)
	  SET_SRC (pat) = gen_rtx_REG (GET_MODE (src), LINK_REG);
	}
    }
  loop_count_setter[0] = NULL_RTX;
  loop_count_setter[GPR_GAP] = NULL_RTX;
  count_life[0] = 0;
  count_life[GPR_GAP] = 0;
  for (insn = first; insn; insn = NEXT_INSN (insn))
    {
      if (GET_CODE (insn) == JUMP_INSN
	  && INSN_CODE (insn) == CODE_FOR_dbra_i)
	{
	  int i = (true_regnum (SET_DEST (XVECEXP (PATTERN (insn), 0, 1)))
		   - LOOP_COUNT0);

	  try_fixup_dbra (loop_count_setter[i], insn,
			  count_life[i]++, label_life[i]);
	  loop_count_setter[i] = NULL_RTX;
	}
      else if (GET_CODE (insn) == INSN
	       && GET_CODE (PATTERN (insn)) == PARALLEL
	       && get_attr_loop_count_set (insn))
	{
	  int n = true_regnum (SET_DEST (XVECEXP (PATTERN (insn), 0, 0)));

	  if (LOOP_COUNT_REGNUM_P (n))
	    {
	      int i = n - LOOP_COUNT0;

	      count_life[i]++;
	      loop_count_setter[i] = insn;
	    }
	}
    }
}

#endif

int
risc16_round_type_align_p (type)
     tree type;
{
  if (TREE_CODE (type) == RECORD_TYPE
      || TREE_CODE (type) == UNION_TYPE
      || TREE_CODE (type) == QUAL_UNION_TYPE)
    {
      tree fields, size;
      int size_sum = 0;

      for (fields = TYPE_FIELDS (type); fields; fields = TREE_CHAIN (fields))
	{
	  size = TYPE_SIZE (TREE_TYPE (fields));
	  if (! size
	      || TREE_CODE (size) != INTEGER_CST
	      || compare_tree_int (size, 4 * BITS_PER_UNIT) > 0)
	    size_sum += 4 * BITS_PER_UNIT;
	  else
	    size_sum += TREE_INT_CST_LOW (size);
	}
      return size_sum >= 4 * BITS_PER_UNIT;
    }
  return 0;
}

/* Like force_operand, but guarantees that VALUE ends up in TARGET.  */
void
force_into (value, target)
     rtx value, target;
{
  value = force_operand (value, target);
  if (! rtx_equal_p (value, target))
    emit_insn (gen_move_insn (target, value));
}
/* END CYGNUS LOCAL -- amylaar/calmrisc16 */


